-- scope stuff
qualityoflife = qualityoflife or {};
local q = qualityoflife;

q.classicheals = q.classicheals or {};
local c = q.classicheals;
c.name = "QoL_ClassicHeals";
c.version = 4; -- check for toc
c.minCoreVersion = 3

QOL_OPTIONS = QOL_OPTIONS or {};
QOL_OPTIONS[c.name] = QOL_OPTIONS[c.name] or {};

-- variables
local frames = {};
local frameName_Target = "CH_TargetFrame";
local framePrefix_Party, framePrefix_RCRF, framePrefix_RPOG = "CH_PartyFrame", "CH_RCRFrame", "CH_RPOGFrame";

-- init
function c:Init()
	if q.initFinished and not c.initFinished and not c.initFailed then
		if q.version >= c.minCoreVersion then
			c:InitOptions();

			c.initFinished = true;
			c:Println(q:GetText("Module loaded."));
		else
			c.initFailed = true;
			c:Println(q:GetText("Error while initialization: QoL Core is outdated (version %s)! Minimum required version: %s", q.version, c.minCoreVersion));
			c:Println(q:GetText("Maybe another of your QoL addon comes with an older version. Please use the latest QoL_Core folder or download the latest version from CurseForge."));
		end
	end
end

-- event handling
local eventFrame;
function c:GetEventFrame()
	eventFrame = eventFrame or q:GetEventFrame();
	return eventFrame;
end
c:GetEventFrame():RegisterEvent("PLAYER_ENTERING_WORLD");
c:GetEventFrame():RegisterEvent("PLAYER_TARGET_CHANGED");
c:GetEventFrame():RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
c:GetEventFrame():HookScript("OnEvent", function(self, event, arg1)
	if event then
		if event == "ADDON_LOADED" and arg1 == c.name then
			c:Init();
		elseif event == "PLAYER_TARGET_CHANGED" then
			if c:GetFrame(frameName_Target) then
				UIFrameFadeOut(c:GetFrame(frameName_Target), 0, 1, 0);
			end
		elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
			c:OnEvent(event, CombatLogGetCurrentEventInfo());
		end
	end
end);

function c:OnEvent(event, ...)
	local timestamp, subevent, _, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags = ...;

	if sourceName == q:GetCharName() and (subevent == "SPELL_HEAL" or subevent == "SPELL_PERIODIC_HEAL") then
		local spellId, spellName, spellSchool, amount, overhealing, absorbed, critical = select(12, ...);
		local parentFrame;

		-- party/raid frame
		local raidId, subGroupId, playerInSubGroupId = c:GetAffectedRaidMemberId(destName);
		if raidId and subGroupId and playerInSubGroupId then
			-- compactraidgroup
			if _G["CompactRaidFrameContainer"]:IsVisible() then
				if CompactRaidFrameManager_GetSetting("KeepGroupsTogether") then
					parentFrame = c:GetOriginalCompactRaidFrame_GKT(subGroupId, playerInSubGroupId);
					if parentFrame and parentFrame:IsVisible() then
						c:DisplayHeal(c:GetFrame(framePrefix_RCRF .. raidId, parentFrame), amount, overhealing, critical);
					end
				else
					-- compactraidframe
					local compactRaidFrameId = c:GetCompactRaidFrameId(destName);
					if compactRaidFrameId then
						parentFrame = c:GetOriginalCompactRaidFrame(compactRaidFrameId);
						if parentFrame and parentFrame:IsVisible() then
							c:DisplayHeal(c:GetFrame(framePrefix_RCRF .. raidId, parentFrame), amount, overhealing, critical);
						end
					end
				end
			end

			-- pulloutgroup
			local pulloutGroupId = c:GetRaidPulloutId(subGroupId);
			if pulloutGroupId then
				parentFrame = c:GetOriginalRaidPulloutGroup(pulloutGroupId, playerInSubGroupId);
				if parentFrame and parentFrame:IsVisible() then
					c:DisplayHeal(c:GetFrame(framePrefix_RPOG .. raidId, parentFrame), amount, overhealing, critical);
				end
			end
		end

		local partyMemberId = c:GetAffectedPartyMemberId(destName);
		if partyMemberId then
			parentFrame = c:GetOriginalPartyFrame(partyMemberId);
			if parentFrame:IsVisible() then
				c:DisplayHeal(c:GetFrame(framePrefix_Party .. partyMemberId, parentFrame), amount, overhealing, critical);
			end
		end

		-- target frame
		if destName == GetUnitName("target") then
			c:DisplayHeal(c:GetFrame(frameName_Target, c:GetOriginalTargetFrame()), amount, overhealing, critical);
		end
	end
end

-- main stuff
function c:DisplayHeal(frame, amount, overhealing, critical)
	if frame then
		local profile, message = c:GetProfile(frame);

		if overhealing > 0 then
			local totalAmount = amount - overhealing;
			if QOL_OPTIONS[c.name][profile][c.OPT_UIATTR_SHOWOVERHEAL] then
				message = "+" .. totalAmount .. " |c" .. q:GetColorHexFromRGBA(QOL_OPTIONS[c.name][profile][c.OPT_UIATTR_FONTCOLOR_OVERHEAL]) .. "(" .. overhealing .. ")";
			else
				if totalAmount > 0 then
					message = "+" .. totalAmount;
				else
					message = "|c" .. q:GetColorHexFromRGBA(QOL_OPTIONS[c.name][profile][c.OPT_UIATTR_FONTCOLOR_OVERHEAL]) .. q:GetText("full");
				end
			end
		else
			message = "+" .. amount;
		end

		c:SetFramePosition(frame);
		c:SetFont(frame, critical);
		frame:SetText(message);

		if critical then
			UIFrameFadeIn(frame, 0.1, 0, 1);
			frame.fadeInfo.finishedFunc = function()
				UIFrameFadeOut(frame, 3, 3, 0);
			end
		else
			UIFrameFadeIn(frame, 0.3, 0, 1);
			frame.fadeInfo.finishedFunc = function()
				UIFrameFadeOut(frame, 1.5, 1.5, 0);
			end
		end
	end
end

function c:GetAffectedPartyMemberId(memberName)
	for playerId, playerName in ipairs(q:GetPartyMembers()) do
		if playerName == memberName then
			return playerId;
		end
	end
end

function c:GetAffectedRaidMemberId(memberName)
	local rm = q:GetRaidMembers();
	local raidId;

	for playerId, playerName in ipairs(rm[0]) do
		if playerName == memberName then
			raidId = playerId;
			break;
		end
	end

	for subGroupId, subGroup in pairs(rm) do
		if subGroupId > 0 then
			for playerInSubGroupId, playerName in ipairs(rm[subGroupId]) do
				if playerName == memberName then
					return raidId, subGroupId, playerInSubGroupId;
				end;
			end;
		end
	end
end

function c:GetCompactRaidFrameId(memberName)
	for n = 1, 40 do
		local compactRaidFrameName = _G["CompactRaidFrame" .. n .. "Name"];
		if compactRaidFrameName and compactRaidFrameName:GetText() == memberName then
			return n;
		end
	end
end

function c:GetRaidPulloutId(subGroupId)
	for n = 1, 8 do
		local raidPulloutName = _G["RaidPullout" .. n .. "Name"];
		if raidPulloutName then
			local pulloutText = raidPulloutName:GetText();
			local textLen = string.len(pulloutText);
			local pulloutGroupId = tonumber(string.sub(pulloutText, textLen, textLen + 1));
			if pulloutGroupId == subGroupId then
				return n;
			end
		end
	end
end

-- misc stuff
function c:GetFrame(name, parent)
	if not frames[name] and parent then
		local newFrame = CreateFrame("Frame", name, parent);
		newFrame:SetWidth(160);
		newFrame:SetHeight(32);
		newFrame:SetFrameStrata("HIGH");
		newFrame.fontString = newFrame:CreateFontString(name .. "_FontString", "ARTWORK", "GameFontNormal");
		newFrame.fontString:SetPoint("CENTER");

		function newFrame.SetText(self, text)
			self.fontString:SetText(text);
		end

		local S = q:GetElvUiSkinModule();
		if S then
			S:HandleFrame(newFrame);
			newFrame:SetBackdropColor(0, 0, 0, 0);
			newFrame:SetBackdropBorderColor(0, 0, 0, 0);
		end

		frames[name] = newFrame;
		frames[name]:Show();
	end
	return frames[name];
end

function c:SetFramePosition(frame)
	local profile = c:GetProfile(frame);
	frame:SetPoint("CENTER", QOL_OPTIONS[c.name][profile][c.OPT_UIATTR_XOFFSET], QOL_OPTIONS[c.name][profile][c.OPT_UIATTR_YOFFSET]);
end

function c:SetFont(frame, critical)
	local profile, size, color = c:GetProfile(frame);

	if critical then
		size = QOL_OPTIONS[c.name][profile][c.OPT_UIATTR_FONTSIZE_CRITICAL];
		color = QOL_OPTIONS[c.name][profile][c.OPT_UIATTR_FONTCOLOR_CRITICAL];
	else
		size = QOL_OPTIONS[c.name][profile][c.OPT_UIATTR_FONTSIZE];
		color = QOL_OPTIONS[c.name][profile][c.OPT_UIATTR_FONTCOLOR];
	end

	if size then
		if profile == c.OPT_UI_RCRF then
			size = size * min(DefaultCompactUnitFrameSetupOptions.height / 36, DefaultCompactUnitFrameSetupOptions.width / 72) -- thx to rulezyx
		end
		frame.fontString:SetFont(frame.fontString:GetFont(), size, "OUTLINE");
	end

	if color then
		frame.fontString:SetTextColor(q:GetColorRBGA(color));
	end
end

function c:GetProfile(frame)
	if frame:GetName() == frameName_Target then
		return c.OPT_UI_TARGET;
	elseif q:StartsWith(frame:GetName(), framePrefix_RPOG) then
		return c.OPT_UI_RPOG;
	elseif q:StartsWith(frame:GetName(), framePrefix_RCRF) then
		return c.OPT_UI_RCRF;
	else
		return c.OPT_UI_PARTY;
	end
end

function c:GetOriginalTargetFrame()
	if ElvUI then
		return _G["ElvUF_Target"];
	end
	return _G["TargetFrame"];
end

function c:GetOriginalPartyFrame(partyMemberId)
	if ElvUI then
		return _G["ElvUF_PartyGroup1UnitButton" .. partyMemberId + 1];
	end
	return _G["PartyMemberFrame" .. partyMemberId];
end

function c:GetOriginalCompactRaidFrame_GKT(subGroupId, playerInSubGroupId)
	if ElvUI then

	end
	return _G["CompactRaidGroup" .. subGroupId .. "Member" .. playerInSubGroupId];
end

function c:GetOriginalCompactRaidFrame(compactRaidFrameId)
	if ElvUI then

	end
	return _G["CompactRaidFrame" .. compactRaidFrameId];
end

function c:GetOriginalRaidPulloutGroup(pulloutGroupId, playerInSubGroupId)
	if ElvUI then

	end
	return _G["RaidPullout" .. pulloutGroupId .. "Button" .. playerInSubGroupId];
end

function c:Println(str)
	if str then
		q:Println(str, "ClassicHeals");
	end
end